/*
 * Copyright 2010 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

/*
 * @test
 * @bug 6754038
 * @summary MethodHandle and MethodType literals
 * @author jrose
 *
 * @compile -source 7 -XDMethodHandleLiterals -target 7 MHLiterals.java
 */

/*
 * Standalone testing:
 * <code>
 * $ cd $MY_REPO_DIR/langtools
 * $ (cd make; make)
 * $ ./dist/bootstrap/bin/javac -XDMethodHandleLiterals -d dist test/tools/javac/meth/MHLiterals.java
 * $ javap -c -classpath dist meth.InvokeMH
 * </code>
 */

package meth;

public class MHLiterals {
    static String lastm;
    static void lastm(String x) { assertEqual(lastm, x); }
    static boolean assertEqual(Object x, Object y) {
        if (x == y || x != null && x.equals(y))  return true;
        System.out.println("*** FAILED assertEqual: "+x+" != "+y);
        Thread.dumpStack();
        return false;
    }

    public void m1(int x, int y) { lastm = "m1"; }
    public void m2() { lastm = "m2()"; }
    public void m2(int x) { lastm = "m2(I)"; }
    public void m2(String x) { lastm = "m2(String)"; }
    public String m2(Object x) { return lastm = "m2(Object)"; }  // different return type
    static void m3() { lastm = "m3()"; }  // static method
    public void m4() { }

    static final int v0_INIT = 1000;
    protected int v0 = v0_INIT;  // inherited into Vars
    public String toString() { return getClass().getSimpleName()+"/v0="+v0; }

    static int nextse, lastse;
    static void lastse(int x) { assertEqual(lastse, x); }
    static <T> T se(T x, int se) { lastse = se; return x; }

    static void staticTest(MHLiterals self) throws Throwable {
        // Unbound and bound forms; ellipsis matching for non-overloads:
        MHLiterals.&m1(int,int).<void>invokeExact(self,1,2); lastm("m1");
        self.&m1(int,int).<void>invokeExact(1,2); lastm("m1");
        self.&m1(...).<void>invokeExact(1,2); lastm("m1");
        // Overloaded method selection:
        MHLiterals.&m2().<void>invokeExact(self); lastm("m2()");
        MHLiterals.&m2(int).<void>invokeExact(self,1); lastm("m2(I)");
        MHLiterals.&m2(String).<void>invokeExact(self,""); lastm("m2(String)");
        MHLiterals.&m2(Object).<String>invokeExact(self,(Object)""); lastm("m2(Object)");
        // Static method selection:
        MHLiterals.&m3().<void>invokeExact(); lastm("m3()");
        se(self,nextse).&m3().<void>invokeExact(); lastm("m3()"); lastse(nextse++);
    }

    void nonStaticTest() throws Throwable {
        // Unbound and bound forms; ellipsis matching for non-overloads:
        MHLiterals.&m1(int,int).<void>invokeExact(this,1,2); lastm("m1");
        this.&m1(int,int).<void>invokeExact(1,2); lastm("m1");
        this.&m1(...).<void>invokeExact(1,2); lastm("m1");
        // Overloaded method selection:
        MHLiterals.&m2().<void>invokeExact(this); lastm("m2()");
        MHLiterals.&m2(int).<void>invokeExact(this,1); lastm("m2(I)");
        MHLiterals.&m2(String).<void>invokeExact(this,""); lastm("m2(String)");
        MHLiterals.&m2(Object).<String>invokeExact(this,(Object)""); lastm("m2(Object)");
        // Static method selection:
        MHLiterals.&m3().<void>invokeExact(); lastm("m3()");
        se(this,nextse).&m3().<void>invokeExact(); lastm("m3()"); lastse(nextse++);
    }

    static void typeTest(MHLiterals self) {
        // Unbound and bound forms; ellipsis matching for non-overloads:
        assertType(MHLiterals.&m1(int,int).type(),  void.class, MHLiterals.class, int.class, int.class);
        assertType(self.&m1(int,int).type(),        void.class,                   int.class, int.class);
        assertType(self.&m1(...).type(),            void.class,                   int.class, int.class);
        // Overloaded method selection:
        assertType(MHLiterals.&m2().type(),         void.class,   MHLiterals.class);
        assertType(MHLiterals.&m2(int).type(),      void.class,   MHLiterals.class, int.class);
        assertType(MHLiterals.&m2(String).type(),   void.class,   MHLiterals.class, String.class);
        assertType(MHLiterals.&m2(Object).type(),   String.class, MHLiterals.class, Object.class);
        // Static method selection:
        assertType(MHLiterals.&m3().type(),             void.class);
        assertType(se(self,nextse).&m3().type(), void.class); lastse(nextse++);
        // Inherited methods must have narrowed receivers:
        assertType(Sub.&m1(...).type(),             void.class, Sub.class, int.class, int.class);
        assertType(Sub.&m3().type(),                void.class);  // static
        assertType(Sub.&m4().type(),                void.class, Sub.class);
    }

    static void assertType(java.dyn.MethodType mt, Class<?> rt, Class<?>... pt) {
        boolean ok = assertEqual(mt.returnType(), rt);
        ok &= assertEqual(mt.parameterCount(), pt.length);
        for (int i = 0; i < pt.length; i++)
            ok &= assertEqual(mt.parameterType(i), pt[i]);
        if (!ok)  System.out.println("*** TYPE: "+mt);
    }

    static class Sub extends MHLiterals {
        //inherit m3, m4, some m2
        public void m1(int x, int y) { lastm = "m1/Sub"; }  // full override
        public void m2(String x) { lastm = "m2(String)/Sub"; }  // partial override

        static final int V0_DELTA = 5000000;
        { v0 += V0_DELTA; }  //inherit int v0

        void nonStaticSubTest() throws Throwable {
            // Unbound and bound forms; ellipsis matching for non-overloads:
            this.&m1(int,int).<void>invokeExact(1,2); lastm("m1/Sub");
            this.&m1(...).<void>invokeExact(1,2); lastm("m1/Sub");
            // invokespecial versions:
            super.&m1(int,int).<void>invokeExact(1,2); lastm("m1");
            super.&m1(...).<void>invokeExact(1,2); lastm("m1");
            // Overloaded method selection, with inheritance:
            Sub.&m2().<void>invokeExact(this); lastm("m2()");
            Sub.&m2(int).<void>invokeExact(this,1); lastm("m2(I)");
            Sub.&m2(String).<void>invokeExact(this,""); lastm("m2(String)/Sub");
            Sub.&m2(Object).<String>invokeExact(this,(Object)""); lastm("m2(Object)");
            // Static method selection:
            Sub.&m3().<void>invokeExact(); lastm("m3()");
            se(this,nextse).&m3().<void>invokeExact(); lastm("m3()"); lastse(nextse++);
        }
    }

    static void staticSubTest(Sub self) throws Throwable {
        // Unbound and bound forms; ellipsis matching for non-overloads:
        self.&m1(int,int).<void>invokeExact(1,2); lastm("m1/Sub");
        self.&m1(...).<void>invokeExact(1,2); lastm("m1/Sub");
        // Overloaded method selection, with inheritance:
        self.&m2().<void>invokeExact(); lastm("m2()");
        self.&m2(int).<void>invokeExact(1); lastm("m2(I)");
        self.&m2(String).<void>invokeExact(""); lastm("m2(String)/Sub");
        self.&m2(Object).<String>invokeExact(""); lastm("m2(Object)");
        // Static method selection:
        Sub.&m3().<void>invokeExact(); lastm("m3()");
        se(self,nextse).&m3().<void>invokeExact(); lastm("m3()"); lastse(nextse++);
    }

    interface Int {
        // for testing with invokeinterface
        void m1(int x, int y);
        void m2();
        void m2(int x);
        void m2(String x);
        String m2(Object x);
        // for testing constant reference:
        String c1 = "c1";
    }

    static class Impl extends MHLiterals implements Int { }

    static void interfaceTest(Int self) throws Throwable {
        // Unbound and bound forms; ellipsis matching for non-overloads:
        Int.&m1(int,int).<void>invokeExact(self,1,2); lastm("m1");
        self.&m1(int,int).<void>invokeExact(1,2); lastm("m1");
        self.&m1(...).<void>invokeExact(1,2); lastm("m1");
        // Overloaded method selection:
        Int.&m2().<void>invokeExact(self); lastm("m2()");
        Int.&m2(int).<void>invokeExact(self,1); lastm("m2(I)");
        Int.&m2(String).<void>invokeExact(self,""); lastm("m2(String)");
        Int.&m2(Object).<String>invokeExact(self,(Object)""); lastm("m2(Object)");
        // Constant value:
        assertEqual( Int.&c1.<String>invokeExact(), Int.c1 );
        assertEqual( se(self,nextse).&c1.<String>invokeExact(), Int.c1 ); lastse(nextse++);
    }

    /*non-static*/ class Nest {
        void nonStaticNestTest() throws Throwable {
            // Unbound and bound forms; ellipsis matching for non-overloads:
            MHLiterals.this.&m1(int,int).<void>invokeExact(1,2); lastm("m1/Sub");
            MHLiterals.this.&m1(...).<void>invokeExact(1,2); lastm("m1/Sub");
            // Overloaded method selection, with inheritance:
            MHLiterals.this.&m2().<void>invokeExact(); lastm("m2()");
            MHLiterals.this.&m2(int).<void>invokeExact(1); lastm("m2(I)");
            MHLiterals.this.&m2(String).<void>invokeExact(""); lastm("m2(String)/Sub");
            MHLiterals.this.&m2(Object).<String>invokeExact(""); lastm("m2(Object)");
            // Static method selection:
            MHLiterals.&m3().<void>invokeExact(); lastm("m3()");
            se(MHLiterals.this,nextse).&m3().<void>invokeExact(); lastm("m3()"); lastse(nextse++);
        }

        // Note: Because all member references must be qualified, there
        // is no need to test unqualified uplevel references like " &m2() ".

        int v4 = 444;
        public String toString() { return MHLiterals.this+"."+super.toString()+",v4="+v4; }
    }

    static String nextstr = "10";
    static void nextstr() { nextstr = ""+(Integer.parseInt(nextstr)+1); }
    static void nextstr(String x) { assertEqual(x, nextstr); nextstr(); }
    static int nextint = 10000;
    static void nextint(int x) { assertEqual(x, nextint++); }

    static class Vars extends MHLiterals {
        String v1 = "garbage";
        static String v2 = "v2";

        public String toString() { return super.toString()+",v1="+v1; }
        Vars(int v0_delta, String v1) { super(v0_delta); this.v1 = v1; }

        void nonStaticGetterTest() throws Throwable {
            // Unbound and bound forms:
            assertEqual( Vars.&v1.<String>invokeExact(this), this.v1 ); 
            assertEqual( this.&v1.<String>invokeExact(), this.v1 );
            assertEqual( Vars.&v2.<String>invokeExact(), Vars.v2 ); 
            assertEqual( se(this,nextse).&v2.<String>invokeExact(), Vars.v2 );  lastse(nextse++);
            // Inheritance:
            assertEqual( Vars.&v0.<int>invokeExact(this), this.v0++ );
            assertEqual( MHLiterals.&v0.<int>invokeExact((MHLiterals)this), this.v0++ );
        }

        void nonStaticSetterTest() throws Throwable {
            // Unbound and bound forms:
            Vars.&=v1.<void>invokeExact(this,nextstr); nextstr(this.v1);
            this.&=v1.<void>invokeExact(nextstr); nextstr(this.v1);
            Vars.&=v2.<void>invokeExact(nextstr); nextstr(Vars.v2); 
            se(this,nextse).&=v2.<void>invokeExact(nextstr); nextstr(Vars.v2);  lastse(nextse++);
            // Inheritance:
            Vars.&=v0.<void>invokeExact(this, nextint); nextint(this.v0);
            MHLiterals.&=v0.<void>invokeExact(this, nextint); nextint(this.v0);
        }
    }

    static void staticGetterTest(Vars self) throws Throwable {
        // Unbound and bound forms:
        assertEqual( Vars.&v1.<String>invokeExact(self), self.v1 ); 
        assertEqual( self.&v1.<String>invokeExact(), self.v1 );
        assertEqual( Vars.&v2.<String>invokeExact(), Vars.v2 ); 
        assertEqual( se(self,nextse).&v2.<String>invokeExact(), Vars.v2 );  lastse(nextse++);
        // Inheritance:
        assertEqual( Vars.&v0.<int>invokeExact(self), self.v0++ );
        assertEqual( MHLiterals.&v0.<int>invokeExact((MHLiterals)self), self.v0++ );
    }

    static void staticSetterTest(Vars self) throws Throwable {
        // Unbound and bound forms:
        Vars.&=v1.<void>invokeExact(self,nextstr); nextstr(self.v1);
        self.&=v1.<void>invokeExact(nextstr); nextstr(self.v1);
        Vars.&=v2.<void>invokeExact(nextstr); nextstr(Vars.v2); 
        se(self,nextse).&=v2.<void>invokeExact(nextstr); nextstr(Vars.v2);  lastse(nextse++);
        // Inheritance:
        Vars.&=v0.<void>invokeExact(self, nextint); nextint(self.v0);
        MHLiterals.&=v0.<void>invokeExact(self, nextint); nextint(self.v0);
    }

    static void varTypeTest(Vars self) {
        // Getters (bound and unbound):
        assertType(Vars.&v1.type(),  String.class, Vars.class);
        assertType(self.&v1.type(),  String.class);
        assertType(Vars.&v2.type(),  String.class);
        assertType(se(self,nextse).&v2.type(), String.class);  lastse(nextse++);
        // Inheritance:
        assertType(Vars.&v0.type(),       int.class, Vars.class);
        assertType(MHLiterals.&v0.type(), int.class, MHLiterals.class);

        // Setters (bound and unbound):
        assertType(Vars.&=v1.type(), void.class, Vars.class, String.class);
        assertType(self.&=v1.type(), void.class,             String.class);
        assertType(Vars.&=v2.type(), void.class,             String.class);
        assertType(se(self,nextse).&=v2.type(), void.class, String.class);  lastse(nextse++);
        // Inheritance:
        assertType(Vars.&=v0.type(),       void.class, Vars.class,       int.class);
        assertType(MHLiterals.&=v0.type(), void.class, MHLiterals.class, int.class);
    }

    // override of hashcode:
    public int hashCode() { return super.hashCode() + 1; }

    void testObjectMethod() throws Throwable {
        int idHash = System.identityHashCode(this);
        assertEqual(super.hashCode(),                      idHash); // invokespecial
        assertEqual(super.&hashCode().<int>invokeExact(),  idHash); // invokespecial
        assertEqual(this.hashCode(),                                   idHash+1);
        assertEqual(this.&hashCode().<int>invokeExact(),               idHash+1);
        assertEqual(MHLiterals.&hashCode().<int>invokeExact(this),     idHash+1);
        assertEqual(Object.&hashCode().<int>invokeExact((Object)this), idHash+1);
        // Check some types also:
        assertType(super.&hashCode().type(),      int.class);
        assertType(this.&hashCode().type(),       int.class);
        assertType(MHLiterals.&hashCode().type(), int.class, MHLiterals.class);
        assertType(Object.&hashCode().type(),     int.class, Object.class);
    }

    MHLiterals(int x) { v0 += x; }
    MHLiterals() { this(0); }

    static void testConstructors() throws Throwable {
        MHLiterals self123 = new MHLiterals(123);
        assertObject(new MHLiterals(), 0);  // can't fail
        assertObject(new MHLiterals&().<MHLiterals>invokeExact(), 0);
        assertType(new MHLiterals&().type(), MHLiterals.class);

        self123 = new MHLiterals&(int).<MHLiterals>invokeExact(123);
        assertObject(self123, 123);
        assertEqual(self123.toString(), new MHLiterals(123).toString());
        assertType(new MHLiterals&(int).type(), MHLiterals.class, int.class);

        assertObject(new Sub(), Sub.V0_DELTA);  // can't fail
        Sub sself = new Sub&().<Sub>invokeExact();
        assertObject(sself, Sub.V0_DELTA);
        assertEqual(sself.toString(), new Sub().toString());
        assertType(new Sub&().type(), Sub.class);

        final int V0_DELTA = 99;
        assertObject(new MHLiterals.Vars(V0_DELTA,"v1"), V0_DELTA);  // can't fail
        Vars vself = new MHLiterals.Vars&(...).<Vars>invokeExact(V0_DELTA,"v1");
        assertObject(vself, V0_DELTA);
        assertEqual(vself.toString(), new MHLiterals.Vars(V0_DELTA,"v1").toString());
        assertType(new MHLiterals.Vars&(...).type(), Vars.class, int.class, String.class);

        String nselfStr = self123.new Nest().toString();
        Nest nself = self123.new Nest&().<Nest>invokeExact();
        assertClass(nself, Nest.class);
        assertEqual(nself.toString(), nselfStr);
        nself = new MHLiterals.Nest&().<Nest>invokeExact(self123);  // explicit outer argument
        assertEqual(nself.toString(), nselfStr);
        nself = new Nest&().<Nest>invokeExact(self123);  // explicit outer argument
        assertEqual(nself.toString(), nselfStr);

        // Constructor for an anonymous class.  Risky, because we don't really know what args it takes.
        assertObject(new Sub&() { { v0 += 333; } }.<Sub>invokeGeneric(), 333);
    }

    static void assertObject(MHLiterals x, int arg) {
        if (!(MHLiterals.class.isAssignableFrom(x.getClass())))
            assertClass(x, MHLiterals.class);
        if (!assertEqual(x.v0, v0_INIT + arg))
            System.out.println("*** OBJECT: "+x);
    }
    static void assertClass(Object x, Class<?> y) {
        boolean ok = assertEqual(x.getClass(), y);
        if (!ok)  System.out.println("*** OBJECT: "+x);
    }

    public static void main(String... av) throws Throwable {
        MHLiterals self = new MHLiterals();
        System.out.println("self="+self);
        staticTest(self);
        self.nonStaticTest();
        typeTest(self);
        self.testObjectMethod();

        Sub sself = new Sub();
        System.out.println("sself="+sself);
        staticSubTest(sself);
        sself.nonStaticSubTest();

        Int iself = new Impl();
        System.out.println("iself="+iself);
        interfaceTest(iself);

        Nest nself = self.new Nest();
        System.out.println("nself="+nself);
        nself.nonStaticNestTest();

        Vars vself = new Vars(0,"v1");
        System.out.println("vself="+vself);
        staticGetterTest(vself);
        vself.nonStaticGetterTest();
        staticSetterTest(vself);
        vself.nonStaticSetterTest();
        varTypeTest(vself);

        testConstructors();
    }
}
